import numpy as np
import sys
from skimage.transform import rescale, warp
from sklearn.metrics import mean_squared_error as mse
from absl import app
from absl import flags


from kernels.hilbert import HilbertGP
from kernels.numeric import Numeric
from placementcriteria.entropy_CT import Entropy_CT, Entropy_CT2


flags.DEFINE_string('criter', "entropy",
                    'entropy/even/random')
flags.DEFINE_string('summarytype', "mean",
                    'mean/max/min')

FLAGS = flags.FLAGS

def rmse(real, predicted):
    return np.sqrt(mse(real, predicted))

def psnr(real, predicted):
    peakval = np.max(real)
    return 10 * np.log10(peakval ** 2 / mse(real, predicted))

def re(real, predicted):
    return np.linalg.norm(real - predicted) / np.linalg.norm(real)


def get_integrals(angle, image, center):
    angle = np.deg2rad(angle)
    cos_a, sin_a = np.cos(angle), np.sin(angle)
    R = np.array([[cos_a, sin_a, -center * (cos_a + sin_a - 1)],
                  [-sin_a, cos_a, -center * (cos_a - sin_a - 1)],
                  [0, 0, 1]])
    rotated_ = warp(image, R, clip=False)
    return rotated_.sum(0)


def compute_integrals(angles, image, center):
    y = []
    for angle in angles:
        y.extend(get_integrals(angle, image, center))
    return np.array(y)

def compute_errors(Eft, image, Xi):
    mean = Eft.reshape(Xi.shape, order='F')
    return rmse(image, mean), psnr(image, mean), re(image, mean)

def fitGP(chosen, A, K, image, center, t, S, basisfunctions, w0, i, filename, Xi):
    Y = compute_integrals(chosen, image, center)
    Eft, Varft, theta = K.approx_2d(np.array(A), Y, S, basisfunctions, w0, optimize=True)
    error = compute_errors(Eft, image, Xi)
    np.save(filename + t + str(i) + "_eft.npy", Eft)
    np.save(filename + t + str(i) + "_chosen.npy", chosen)

    return error

def entropy(angles, K, maxsensors, basisfunctions, w0, filename, image, center, S,  Xi):
    t = FLAGS.summarytype
    print(t)
    A = []
    chosen = []
    #criterion = Entropy_CT(t, 8., 0.1, 0.1, m=basisfunctions)
    criterion = Entropy_CT2(t, K)
    newlines, chosen_, angles = criterion.initialize(angles)
    A.extend(newlines)
    chosen.append(chosen_)

    for i in range(maxsensors - 1):
        print(i)
        newlines, chosen_, angles = criterion.get_nextlines(A, angles)
        print(chosen_)
        sys.stdout.flush()
        A.extend(newlines)
        chosen.append(chosen_)
        np.save(filename + t + str(i) + "_chosen.npy", chosen)
        print(chosen)

def random(angles, K, maxsensors, basisfunctions, w0, filename, image, center, S,  Xi):
    t = 'random'
    A = []
    errors = []
    chosen = np.random.choice(angles, maxsensors, replace=False)
    criterion = Entropy_CT(t, 8., 0.1, 0.1, m=basisfunctions)

    for i, angle in enumerate(chosen):
        print(i)
        newlines = criterion.get_newlines(angle)
        A.extend(newlines)

        if i in [1,4,7,8,9]:
            error = fitGP(chosen[:i + 1], A, K, image, center, t, S, basisfunctions, w0, i, filename,  Xi)
            errors.append(error)
            np.save(filename + t + i + "_errors.npy", errors)

    np.save(filename + t + "_errors.npy", errors)

def even(angles, K, maxsensors, basisfunctions, w0, filename, image, center, S,  Xi):
    t = 'even'
    errors = []
    criterion = Entropy_CT(t, 8., 0.1, 0.1, m=basisfunctions)
    for i in range(1, maxsensors):
        angles = np.linspace(0., 180., i, endpoint=False)
        A = []
        for angle in angles:
            newlines = criterion.get_newlines(angle)
            A.extend(newlines)

        if i in [2,5,8,9,10]:
            error = fitGP(angles, A, K, image, center, t, S, basisfunctions, w0, i, filename,  Xi)
            errors.append(error)
            np.save(filename + t + i + "_errors.npy", errors)

    np.save(filename + t + "_errors.npy", errors)


def main(argv):

    image = np.genfromtxt('data/CT/lungsegment.txt', delimiter=',')
    image = (image - np.min(image)) / np.max(image)
    image = rescale(image, scale=0.25, mode='reflect', multichannel=False)

    # Test Data
    n = image.shape[0]
    [Xi, Xj] = np.meshgrid(np.linspace(0,image.shape[0], n), np.linspace(0, image.shape[0], n))
    S = np.vstack((Xi.ravel(), Xj.ravel())).T

    maxsensors = 10
    basisfunctions = 100
    filename = "data/CT/ct_num"
    center = 63.5
    K = HilbertGP('Matern', v=1)
    w0 = np.log(np.array([8., 0.1, 0.1]))
    angles = np.linspace(0., 180., 18, endpoint=False)

    K = Numeric(lengthscale=8., variance=0.1, noise=0.1, kerneltype='RBF', period=0, samples=10)
    criter = FLAGS.criter

    if criter == 'entropy':
        print('entropy')
        sys.stdout.flush()
        entropy(angles, K, maxsensors, basisfunctions, w0, filename, image, center, S,  Xi)

    elif criter == 'random':
        print("random")
        sys.stdout.flush()
        random(angles, K, maxsensors, basisfunctions, w0, filename, image, center, S,  Xi)

    elif criter == 'even':
        print("even")
        sys.stdout.flush()
        even(angles, K, maxsensors, basisfunctions, w0, filename, image, center, S,  Xi)



if __name__ == "__main__":
    app.run(main)

