import numpy as np
from .stationary import Stationary

class RQ(Stationary):
    def __init__(self, variance=1., lengthscale=1., noise=1.):
        super().__init__(variance, lengthscale, noise)
        self.power = 2

    def kernel_vector(self, x, xnew):
        r = self.unscaled_dist(x, xnew)
        r2 = np.square(r)
        return (1 + r2 / (2 * self.power)) ** (-self.power)

    def analytic_parallel(self, a, b, f, p):
        A = -b ** 2 + 4 * (4 + a) * f

        def term1(s, t):
            val = np.arctan((b - 2 * f * s + 2 * f * p * t) / np.sqrt(A))
            return 2 * f * s * val

        def term2(s, t):
            val = np.arctan((-b + 2 * f * (s - p * t)) / np.sqrt(A))
            return (b + 2 * f * p * t) * val

        def both(s, t):
            return 32 * (term1(s, t) + term2(s, t)) / (np.power(A, (3 / 2)) * p)

        return both(1, 1) + both(0, 0) - both(1, 0) - both(0, 1)

    def analytic_same_startpoint(self, a, b, c, d):
        w2_norm = c ** 2 + d ** 2
        w1_norm = a ** 2 + b ** 2
        Bt = np.sqrt(4 * w1_norm * self.lengthscale ** 2 + (a * d - c * b) ** 2)
        Bs = np.sqrt(4 * w2_norm * self.lengthscale ** 2 + (a * d - c * b) ** 2)
        wv = a * c + b * d

        term1 = np.arctan((-wv + w1_norm) / Bt) / Bt
        term2 = np.arctan((-wv + w2_norm) / Bs) / Bs
        term3 = np.arctan((-wv) / Bs) / Bs
        term4 = np.arctan((-wv) / Bt) / Bt

        return 2 * self.lengthscale ** 2 * (term1 + term2 - term3 - term4)

    def analytic(self, a, b, c, d, e, f):
        div = -b * c * d + (4 + a) * d ** 2 + b ** 2 * e + (c ** 2 - 4 * (4 + a) * e) * f

        AS0 = np.sqrt(-c ** 2 + 4 * e * (4 + a))
        AS1 = np.sqrt(-c ** 2 + 2 * c * d - d ** 2 + 4 * e * (4 + a - b + f))
        BS0 = c * d - 2 * b * e
        BS1 = c * d - 2 * b * e - d ** 2 + 4 * e * f

        AT0 = np.sqrt(-b ** 2 + 4 * f * (4 + a))
        AT1 = np.sqrt(-b ** 2 - 2 * b * d - d ** 2 + 4 * f * (4 + a + c + e))
        BT0 = b * d - 2 * c * f
        BT1 = b * d - 2 * c * f + d ** 2 - 4 * e * f

        term1 = 16 * (
        -BS1 * np.arctan((c - d + 2 * e) / AS1) / AS1 + BT1 * np.arctan((-b + 2 * f - d) / AT1) / AT1) / div
        term2 = -16 * (-BS0 * np.arctan((c + 2 * e) / AS0) / AS0 + BT1 * np.arctan((-b - d) / AT1) / AT1) / div
        term3 = -16 * (-BS1 * np.arctan((c - d) / AS1) / AS1 + BT0 * np.arctan((-b + 2 * f) / AT0) / AT0) / div
        term4 = 16 * (-BS0 * np.arctan(c / AS0) / AS0 + BT0 * np.arctan(-b / AT0) / AT0) / div

        return term1 + term2 + term3 + term4

    def doubleintegralkernel(self, X1, X2):
        p1 = X1[0]
        p2 = X2[0]
        V = np.identity(2) * 1 / (self.lengthscale ** 2)
        u = p1 - p2

        w1 = np.array(self.getw(X1[0], X1[1]))
        w2 = np.array(self.getw(X2[0], X2[1]))
        w1_norm = np.linalg.norm(w1)
        w2_norm = np.linalg.norm(w2)

        if np.abs(w1_norm) < 1e-6 and np.abs(w2_norm) < 1e-6:
            return 0

        a = u.T @ V @ u
        b = 2 * u.T @ V @ w2
        c = 2 * u.T @ V @ w1
        d = 2 * w2.T @ V @ w1
        e = w1.T @ V @ w1
        f = w2.T @ V @ w2

        # p1 - p2 = 0 -> p1 = p2
        # Lähtevät samasta pisteestä. Yksinkertaistuu simple formulationiin
        if a == 0:
            return w1_norm * w2_norm * self.analytic_same_startpoint(w1[0], w1[1], w2[0], w2[1])

        if np.abs(np.cross(w1, w2)) < 1e-8:
            p = w1_norm / w2_norm
            if not np.allclose(w1, p * w2):
                p = -p
            return w1_norm * w2_norm * self.analytic_parallel(a, b, f, p)

        return w1_norm * w2_norm * self.analytic(a, b, c, d, e, f)

    def integralkernel(self, X, z):
        w = np.array(self.getw(X[0], X[1]))
        v = X[0] - z
        w_norm = np.linalg.norm(w)

        if w_norm == 0:
            return 0

        V = np.identity(2) * 1 / (self.lengthscale ** 2)

        c1 = w.T @ V @ w
        c2 = 2 * w.T @ V @ v
        c3 = v.T @ V @ v

        A = 4 * c1 * (c3 + 4) - c2 ** 2
        term1 = (2 * c1 + c2) / (A * (c1 + c2 + c3 + 4))
        term2 = 4 * c1 * np.arctan((2 * c1 + c2) / np.sqrt(A)) / A ** (3 / 2)
        term3 = c2 / (A * (c3 + 4))
        term4 = 4 * c1 * np.arctan(c2 / np.sqrt(A)) / A ** (3 / 2)

        return w_norm * 16 * ((term1 + term2) - (term3 + term4))