import numpy as np
import math
import scipy.optimize as op

class HilbertGP():

    """
    This code is adapted for line integrals from the original matlab code available at
    https://github.com/AaltoML/hilbert-gp
    The original paper corresponding to that code is available at
    https://arxiv.org/pdf/1809.03779.pdf
    """

    def __init__(self, kernel, v=1.):
        self.kernel = kernel
        self.v = v

    def optimize_hyperparams(self, w, y, lambda_, Phiy, PhiPhi, S, dS):
        w = np.squeeze(w)
        # Extract parameters
        lengthScale = np.exp(w[0])
        magnSigma2 = np.exp(w[1])
        sigma2 = np.exp(w[2])

        # Number of n=observations and m=basis functions
        n = np.size(y)
        m = np.shape(Phiy)[0]

        k = S(np.sqrt(lambda_), lengthScale, magnSigma2) + 1e-32
        L = np.linalg.cholesky(PhiPhi + np.eye(len(PhiPhi)) * sigma2 / k)

        # Evaluate all parts
        v = np.linalg.inv(L) @ Phiy
        yiQy = (y.T @ y - v.T @ v) / sigma2
        logdetQ = (n - m) * np.log(sigma2) + np.sum(np.log(k)) + 2 * np.sum(np.log(np.diag(L)))

        # Return approx. negative log marginal likelihood
        e = .5 * yiQy + .5 * logdetQ + .5 * n * np.log(2 * np.pi)

        # Precalculate
        vv = np.linalg.inv(L.T) @ v
        LLk = np.linalg.inv(L.T) @ (np.linalg.inv(L) @ np.diag(1. / k))

        # For the covariance function hyperparameters
        eg = np.zeros(w.shape);
        for j in range(len(w) - 1):

            # Evaluate the partial derivative
            if j == 0:
                dk = dS[0](np.sqrt(lambda_), lengthScale, magnSigma2)
            else:
                dk = dS[1](np.sqrt(lambda_), lengthScale, magnSigma2)

            # Evaluate parts
            dlogdetQ = np.sum(dk / k) - sigma2 * np.sum(np.diag(LLk) * dk / k)
            dyiQy = -vv.T @ np.diag(dk / k ** 2) @ vv

            # The partial derivative
            eg[j] = .5 * dlogdetQ + .5 * dyiQy
            # Account for the log-transformed values
            eg[j] = np.exp(w[j]) * eg[j]

        # Evaluate parts
        dlogdetQ = (n - m) / sigma2 + np.sum(np.diag(LLk))
        dyiQy = vv.T @ np.diag(1. / k) @ vv / sigma2 - yiQy / sigma2

        # For the measurement noise
        eg[-1] = .5 * dlogdetQ + .5 * dyiQy

        # Account for the log-transformed values
        eg[-1] = np.exp(w[-1]) * eg[-1]

        return e, eg

    def approx_2d(self, x, y, xt, m, w0, optimize="True"):

        x = np.squeeze(x)
        d = x.ndim

        Lt = np.array([4 / 3 * np.max(np.abs(x)), 4 / 3 * np.max(np.abs(x))])
        NN = np.linspace(1, m, m)

        def eigenval(n):
            m = len(n)
            lambd = np.zeros(m * m)
            phi1 = np.zeros(m)
            phi2 = np.zeros(m)
            for i1 in range(m):
                for i2 in range(m):
                    i = i1 + m * (i2 - 1)
                    phi1[i1] = (n[i1] * np.pi / (2 * Lt[0]))
                    phi2[i2] = (n[i2] * np.pi / (2 * Lt[1]))
                    lambd[i] = (phi1[i1] ** 2 + phi2[i2] ** 2)
            return lambd, np.array([phi1, phi2])

        def basisfunc(j, x, phi1, phi2):
            return (Lt[0] * Lt[1]) ** (-1 / 2) * np.sin(phi1 * (x[0] + Lt[0])) * np.sin(phi2 * (x[1] + Lt[1]))

        def getw(p1, p2):
            return [p2[0] - p1[0], p2[1] - p1[1]]

        def basisfunc_analytic(j, x, phi1, phi2):
            w = getw(x[0], x[1])
            alpha = phi1 * (w[0])
            gamma = phi2 * (w[1])

            if (alpha == gamma or alpha == - gamma):
                return basisfunc_numeric(j, x, phi1, phi2)

            def term1(t):
                return np.sin(phi1 * (w[0] * t + x[0][0] + Lt[0]) - phi2 * (w[1] * t + x[0][1] + Lt[1])) \
                       / ((gamma - alpha))

            def term2(t):
                return np.sin(phi2 * (w[1] * t + x[0][1] + Lt[1]) + phi1 * (w[0] * t + x[0][0] + Lt[0])) \
                       / ((alpha + gamma))

            multiplier = 1 / (2 * np.sqrt(Lt[0] * Lt[1]))
            return np.linalg.norm(w) * multiplier * (-term1(1) - term2(1) + term1(0) + term2(0))

        def basisfunc_numeric(j, x, phi1, phi2):
            samples = 30
            w = getw(x[0], x[1])
            w_norm = np.linalg.norm(w)
            t = np.linspace(0, 1, samples)[:, None]
            integral = 0
            for i in range(samples):
                x_ = w * t[i] + x[0]
                integral += (Lt[0] * Lt[1]) ** (-1 / 2) * np.sin(phi1 * (x_[0] + Lt[0])) * np.sin(
                    phi2 * (x_[1] + Lt[1]))
            return 1 / samples * w_norm * integral

        def eigenfun(n, x, phi):
            phi1 = phi[0]
            phi2 = phi[1]
            vals = np.zeros((len(x), len(n) * len(n)))
            for i1 in range(len(phi1)):
                for i2 in range(len(phi2)):
                    i = i1 + m * (i2 - 1)
                    for j in range(len(x)):
                        if x.ndim == 2:
                            vals[j, i] = basisfunc(i + 1, x[j], phi1[i1], phi2[i2])
                        elif x.ndim == 3:
                            vals[j, i] = basisfunc_analytic(i + 1, x[j], phi1[i1], phi2[i2])
                        else:
                            raise Exception('Dimensionality of x is wrong')
            return vals

        def S(w, l, magnSigma2):
            if self.kernel == 'RBF':
                return magnSigma2 * np.sqrt(2 * np.pi) ** (d / 2) * lengthScale ** (d) * np.exp(
                    -0.5 * w * w.T * lengthScale ** 2)
            elif self.kernel == 'Matern':
                v = self.v
                return magnSigma2 * ((2 ** d * np.pi ** (d / 2) * math.gamma(v + d / 2) * (2 * v) ** v) / \
                                     (math.gamma(v) * l ** (2 * v))) * ((2 * v) / l ** 2 + w * w.T) ** -(v + d / 2)
            else:
                raise Exception('Kernel not implemented. Try RBF or Matern.')

        def dS_l(w, l, magnSigma2):
            if self.kernel == 'RBF':
                return (2 * np.pi) ** (d / 2) * magnSigma2 * l ** (d - 1) * np.exp(-0.5 * w * w.T * l ** 2) * (
                d - l ** 2 * w * w.T)
            else:
                v = self.v
                return -2 * magnSigma2 * (
                (2 ** d * np.pi ** (d / 2) * math.gamma(v + d / 2) * (2 * v) ** v) / (math.gamma(v))) \
                       * v * l ** (-2 * v - 3) * (l ** 2 * w * w.T + 2 * v + 2)

        def dS_s(w, lengthScale, magnSigma2):
            return S(w, lengthScale, magnSigma2) / magnSigma2

        # Eigenfunctions
        lambda_, phi = eigenval(NN)
        Phit = eigenfun(NN, xt, phi)
        Phi = eigenfun(NN, x, phi)
        PhiPhi = Phi.T @ Phi
        Phiy = Phi.T @ y

        if optimize == True:
            print("Optimizing")
            Result = op.minimize(fun=self.optimize_hyperparams,
                                 x0=w0,
                                 args=(y, lambda_, Phiy, PhiPhi, S, [dS_l, dS_s]),
                                 jac=True,
                                 tol=1e-3);
            w = np.exp(Result.x)
        else:
            w = np.exp(w0)

        lengthScale = w[0]
        magnSigma2 = w[1]
        sigma2 = w[2]

        # Solve GP with optimized hyperparameters
        # return predictive mean and variance
        k = S(np.sqrt(lambda_), lengthScale, magnSigma2) + 1e-32
        L = np.linalg.cholesky(PhiPhi + np.eye(len(PhiPhi)) * sigma2 / k)
        Eft = Phit @ (np.linalg.inv(L.T) @ (np.linalg.inv(L) @ Phiy))
        Varft = sigma2 * np.sum(np.dot(Phit, np.linalg.pinv(L.T)) ** 2, axis=1)

        return [Eft, Varft, w]

    def fit_GP(self, X, S, Y, m=20, optimize=True, w0=np.array([5., 1., 1.])):
        w0 = np.log(w0)
        Eft, Varft, theta = self.approx_2d(X, Y, S, m, w0, optimize=optimize)
        return Eft, Varft, theta
