import numpy as np
from .stationary import Stationary

class Numeric(Stationary):
    """
    Numerically computing kernel function for GPs with line integral observations.
    
    Implemented kernel functions: Matern32, Matern52, RBF, RQ
    
    """
    def __init__(self, kerneltype, variance=1., lengthscale=1., noise=1., samples=10, period=0):
        super().__init__(variance, lengthscale, noise)
        self.samples = samples
        self.kerneltype = kerneltype
        self.period = period
        #
        # Computes the kernel function between x and xnew
        #
        # If period is given for pipe geometry, the points are scaled
        # Assumes that the x-axis for the mapped area is [-period, period]

    def kernel_vector(self, x, xnew):
        def scale(point, period):
            return point + period

        if self.period == 0:
            r = self.unscaled_dist(x, xnew)
        else:
            x_plus = np.array([scale(x[:, 0], self.period), x[:, 1]]).T
            x_minus = np.array([scale(x[:, 0], -self.period), x[:, 1]]).T
            x_plus2 = np.array([scale(x[:, 0], 2 * self.period), x[:, 1]]).T
            x_minus2 = np.array([scale(x[:, 0], -2 * self.period), x[:, 1]]).T
            r = self.unscaled_dist(x, xnew)
            r_plus = self.unscaled_dist(x_plus, xnew)
            r_minus = self.unscaled_dist(x_minus, xnew)
            r_plus2 = self.unscaled_dist(x_plus2, xnew)
            r_minus2 = self.unscaled_dist(x_minus2, xnew)
            r = np.minimum(np.minimum(np.minimum(np.minimum(r, r_minus2), r_plus2), r_plus), r_minus)
        if self.kerneltype == 'RBF':
            return self.variance * np.exp(-0.5 * r ** 2)
        elif self.kerneltype == "RQ":
            power = 2
            r2 = np.square(r)
            return self.variance * np.exp(-power * np.log1p(r2 / 2.))
        elif self.kerneltype == 'Matern32':
            return self.variance * (1. + np.sqrt(3.) * r) * np.exp(-np.sqrt(3.) * r)
        elif self.kerneltype == 'Matern52':
            return self.variance * (1 + np.sqrt(5.) * r + 5. / 3 * r ** 2) * np.exp(-np.sqrt(5.) * r)
        else:
            Exception('Kernel not implemented. Try RBF, RQ, Matern32 or Matern52.')

    # Coputing the kernel between a line and a point (one integral)
    def integralkernel(self, X, Xnew):
        t = np.linspace(0, 1, self.samples)[:, None]
        w = self.getw(X[0], X[1])
        w_norm = np.linalg.norm(w)
        kern = []
        for i in range(self.samples):
            kern.append(self.pointkernel(w * t[i] + X[0], Xnew) * w_norm)
        kern = np.asarray(kern)
        return 1 / self.samples * np.sum(kern)

    # Computing the kernel between lines (double integral)
    def doubleintegralkernel(self, X1, X2):
        t = np.linspace(0, 1, self.samples)[:, None]
        kern = []
        w1 = self.getw(X1[0], X1[1])
        w2 = self.getw(X2[0], X2[1])
        w1_norm = np.linalg.norm(w1)
        w2_norm = np.linalg.norm(w2)
        for h in range(self.samples):
            for k in range(self.samples):
                kern.append(self.pointkernel(w1 * t[h] + X1[0], w2 * t[k] + X2[0]) * w1_norm * w2_norm)
        kern = np.asarray(kern)
        return 1 / self.samples ** 2 * np.sum(kern)