import numpy as np
import math

class Entropy():
    def __init__(self, summarytype):
        self.summarytype = summarytype
        self.decimals = 5

    def entropy(self, sig):
        if sig.ndim > 1:
            sig = np.diagonal(sig)
        return 0.5 * np.log(2 * np.pi * np.exp(1) * sig)

    def score(self, entropy):
        if self.summarytype == 'mean':
            s = np.mean(entropy)
        if self.summarytype == 'min':
            s = np.min(entropy)
        if self.summarytype == 'max':
            s = np.max(entropy)
        return s

    def scores(self, Kaa_inv, chosenlines, sensors, chosensensors):
        scores = []
        lines = []
        for x in sensors:
            newlines = self.get_newlines([x], chosensensors)
            sig = self.sigma(newlines, chosenlines, Kaa_inv)
            s = self.score(self.entropy(sig))
            scores.append(s)
            lines.append(newlines)
        scores = np.round(np.array(scores), decimals=self.decimals)
        return scores, np.array(lines)

    def scores2(self, Kaa_inv, chosenlines, sensors, chosensensors):
        scores = []
        lines = []
        for x in sensors:
            newlines = self.get_newlines(chosensensors, [x])
            sig = self.sigma(newlines, chosenlines, Kaa_inv)
            s = self.score(self.entropy(sig))
            scores.append(s)
            lines.append(newlines)
        scores = np.round(np.array(scores), decimals=self.decimals)
        return scores, np.array(lines)

    def sigma(self, lines, chosensenlines):
        raise NotImplementedError("implement to use this class")

    def get_newlines(self, sensor, chosensensors):
        raise NotImplementedError("implement to use this class")


class Entropy_numeric(Entropy):
    def __init__(self, summarytype, kernel):
        super().__init__(summarytype)
        self.K = kernel

    def sigma(self, S, X, Kxx_inv):
        sigmas = []
        for S_ in S:
            Kxs = self.K.kernel(X, [S_])
            Kss = self.K.kernel([S_], [S_])
            Cov = Kss - Kxs @ Kxx_inv @ Kxs.T
            sigmas.append(np.squeeze(Cov))
        return np.array(sigmas)

    def K_inverse(self, A):
        Kaa = self.K.kernel(A, A) + self.K.noise * np.eye(len(A))
        return np.linalg.inv(Kaa)


class Entropy_BFE(Entropy):
    def __init__(self, summarytype, lengthscale, magnsigma, sigma2, m=44):
        super().__init__(summarytype)
        self.lengthScale = lengthscale
        self.magnSigma2 = magnsigma
        self.sigma2 = sigma2

        self.d = 2
        self.Lt = np.array([128, 128])
        self.m = m
        self.NN = np.linspace(1, m, m)
        self.lambda_, self.phi = self.eigenval(self.NN)

    def sigma(self, xt, L_inv):
        sigmas = []
        for x in xt:
            Phit = self.eigenfun(self.NN, np.array([x]), self.phi)
            Varft = self.sigma2 * np.sum(np.dot(Phit, L_inv) ** 2, axis=1)
            sigmas.append(np.squeeze(Varft))
        return np.array(sigmas)

    def K_inverse(self, x):
        Phi = self.eigenfun(self.NN, x, self.phi)
        PhiPhi = Phi.T @ Phi
        k = self.S(np.sqrt(self.lambda_), self.lengthScale, self.magnSigma2) + 1e-32
        L = np.linalg.cholesky(PhiPhi + np.eye(len(PhiPhi)) * self.sigma2 / k)
        return np.linalg.pinv(L.T)



    def getw(self, p1, p2):
        return [p2[0] - p1[0], p2[1] - p1[1]]

    def eigenval(self, n):
        m = len(n)
        lambd = np.zeros(m * m)
        phi1 = np.zeros(m)
        phi2 = np.zeros(m)
        for i1 in range(m):
            for i2 in range(m):
                i = i1 + m * (i2 - 1)
                phi1[i1] = (n[i1] * np.pi / (2 * self.Lt[0]))
                phi2[i2] = (n[i2] * np.pi / (2 * self.Lt[1]))
                lambd[i] = (phi1[i1] ** 2 + phi2[i2] ** 2)
        return lambd, np.array([phi1, phi2])

    def basisfunc(self, j, x, phi1, phi2):
        return (self.Lt[0] * self.Lt[1]) ** (-1 / 2) * np.sin(phi1 * (x[0] + self.Lt[0])) * np.sin(
            phi2 * (x[1] + self.Lt[1]))

    def basisfunc_analytic(self, j, x, phi1, phi2):
        w = self.getw(x[0], x[1])
        alpha = phi1 * (w[0])
        gamma = phi2 * (w[1])

        if (alpha == gamma or alpha == - gamma):
            return self.basisfunc_numeric(j, x, phi1, phi2)

        def term1(t):
            return np.sin(phi1 * (w[0] * t + x[0][0] + self.Lt[0]) - phi2 * (w[1] * t + x[0][1] + self.Lt[1])) \
                   / ((gamma - alpha))

        def term2(t):
            return np.sin(phi2 * (w[1] * t + x[0][1] + self.Lt[1]) + phi1 * (w[0] * t + x[0][0] + self.Lt[0])) \
                   / ((alpha + gamma))

        multiplier = 1 / (2 * np.sqrt(self.Lt[0] * self.Lt[1]))
        return np.linalg.norm(w) * multiplier * (-term1(1) - term2(1) + term1(0) + term2(0))

    def basisfunc_numeric(self, j, x, phi1, phi2):
        samples = 30
        w = self.getw(x[0], x[1])
        w_norm = np.linalg.norm(w)
        t = np.linspace(0, 1, samples)[:, None]
        integral = 0
        for i in range(samples):
            x_ = w * t[i] + x[0]
            integral += (self.Lt[0] * self.Lt[1]) ** (-1 / 2) * np.sin(phi1 * (x_[0] + self.Lt[0])) * np.sin(
                phi2 * (x_[1] + self.Lt[1]))
        return 1 / samples * w_norm * integral

    def eigenfun(self, n, x, phi):
        phi1 = phi[0]
        phi2 = phi[1]
        vals = np.zeros((len(x), len(n) * len(n)))
        for i1 in range(len(phi1)):
            for i2 in range(len(phi2)):
                i = i1 + self.m * (i2 - 1)
                for j in range(len(x)):
                    if x.ndim == 2:
                        vals[j, i] = self.basisfunc(i + 1, x[j], phi1[i1], phi2[i2])
                    elif x.ndim == 3:
                        vals[j, i] = self.basisfunc_analytic(i + 1, x[j], phi1[i1], phi2[i2])
                    else:
                        raise Exception('Dimensionality of x is wrong')
        return vals

    def S(self, w, l, magnSigma2):
        v = 1.
        d = 2
        return magnSigma2 * ((2 ** d * np.pi ** (d / 2) * math.gamma(v + d / 2) * (2 * v) ** v) / \
                             (math.gamma(v) * l ** (2 * v))) * ((2 * v) / l ** 2 + w * w.T) ** -(v + d / 2)