import numpy as np
from sklearn.metrics import mean_squared_error as mse
import sys
from absl import app
from absl import flags

from kernels.numeric import Numeric
from placementcriteria.entropy_ultrasound import Entropy_ultrasound
from placementcriteria.mi_ultrasound import MI_ultrasound


flags.DEFINE_string('criter', "entropy",
                    'entropy/mi')
flags.DEFINE_string('summarytype', "max",
                    'min/mean/max')
flags.DEFINE_integer('xgrid', "5",
                    'gridsize')
flags.DEFINE_integer('ygrid', "4",
                    'gridsize')
flags.DEFINE_integer('maxsensors', "16",
                    'gridsize')
FLAGS = flags.FLAGS


def error_rmse(estimated, true):
    return np.sqrt(mse(estimated, true))


def error_psnr(predicted, real):
    peakval = np.max(real)
    return 10 * np.log10(peakval ** 2 / mse(real, predicted))


def error_re(real, predicted):
    return np.linalg.norm(real - predicted) / np.linalg.norm(real)


def error(estimated, true):
    return error_rmse(estimated, true)


def getw(p1, p2):
    return [p2[0] - p1[0], p2[1] - p1[1]]


def scale(point):
    period = 2*np.pi*35
    if point[0] > period / 2:
        point = point - [period, 0]
    if point[0] < -period / 2:
        point = point + [period, 0]
    return point


def sample_line(point1, point2, samples):
    p = point1
    w = getw(point1, point2)
    s = np.linspace(0, 1, samples)[:, None]
    points = w * s + p
    points = np.array([scale(x) for x in points])
    return points


def get_integralvalues_for_function(X, f):
    linenoise = 0.001

    def F_orig(S):
        x = S[:, 0]
        y = S[:, 1]
        return value(x, y, f)

    def r(t, p1, p2):
        return np.array([(1 - t) * p1[0] + t * p2[0],
                         (1 - t) * p1[1] + t * p2[1]])

    def delta(point1, point2):
        return np.sqrt((point2[0] - point1[0]) ** 2
                       + (point2[1] - point1[1]) ** 2)

    integrals = []
    for i in range(len(X)):
        line = X[i]
        point1 = line[0]
        point2 = line[1]
        t = np.linspace(0., 1., 100)
        t_diff = np.ediff1d(t, to_begin=0)
        res = 0
        for i in range(1, len(t)):
            res += F_orig(r(np.reshape(t[i], (1,)), point1, point2).T) * delta(r(t[i], point1, point2),
                                                                               r(t[i - 1], point1, point2))
        integrals.append(np.sum(res))

    intgrl = np.array(integrals)
    intgrl = intgrl #+ np.random.normal(0, linenoise, len(integrals))
    return intgrl

def compute_error(X, funcs, K, itr, t, filename, S, Xi, Xj):
    Kxx = K.kernel(X, X) + K.noise * np.eye(len(X))
    Kxx_inv = np.linalg.inv(Kxx)
    Kxs = K.kernel(X, S)

    errors = []
    for m in range(len(funcs)):
        Y = get_integralvalues_for_function(X, funcs[m])[:, None]
        Y_real = np.ones(Xi.shape)
        for i in range(Xi.shape[0]):
            for j in range(Xi.shape[1]):
                Y_real[i,j] = value(Xi[i,j], Xj[i,j], funcs[m])
        test_y = np.array(Y_real)
        mean = Kxs @ Kxx_inv @ Y
        mean = mean.reshape(Xi.shape)
        np.save(filename + t + str(itr) + str(m) + "_mean.npy", mean)
        np.save(filename + t + str(itr) + str(m) + "_real.npy", test_y)
        rms = error(test_y, mean)
        errors.append(rms)
    return np.array(errors)

def entropy(t, V_start, V_end,  functions, S, Xi, Xj, filename, maxsensors):
    K = Numeric(lengthscale=15., variance=0.5, noise=0.001, kerneltype='Matern32', period=period, samples=10)

    rms_all = []
    S_points = V_start
    E_points = V_end
    criterion = Entropy_ultrasound(t, K, second=True)
    A, S_points, E_points, starts, ends = criterion.initialize(S_points, E_points)

    for i in range(maxsensors - 1):
        print(i)
        newlines, S_points, E_points, starts, ends = criterion.get_nextlines(A, S_points, E_points, starts, ends)
        A.extend(newlines)
        np.save(filename + t + str(i) + "_chosen.npy", ends)
        np.save(filename + t + str(i) + "_chosenstarts.npy", starts)
        np.save(filename + t + str(i) + "_lines.npy", A)
        np.save(filename + t + str(i) + "_newlines.npy", newlines)
        rms_all.append(compute_error(A, functions, K, i, t, filename, S, Xi, Xj))


def mi(t, S_points, E_points,  functions, S, Xi, Xj, filename, maxsensors):
    K = Numeric(lengthscale=15., variance=0.5, noise=0.001, kerneltype='Matern32', period=period, samples=10)
    criterion = MI_ultrasound(t, K, second=True)
    lines, S_points, E_points, starts, ends = criterion.initialize(S_points, E_points)
    errors = []
    scores_e = np.ones(len(E_points)) * 100000
    scores_updated_e = np.zeros(len(E_points))
    scores_s = np.ones(len(S_points)) * 100000
    scores_updated_s = np.zeros(len(S_points))

    for i in range(maxsensors - 1):
        print(i)
        sys.stdout.flush()
        newlines, S_points, E_points, starts, ends = criterion.get_nextlines(lines, S_points, E_points, starts, ends,
                                                                             scores_updated_s, scores_s,
                                                                             scores_updated_e, scores_e)


        lines.extend(newlines)
        chosen.append(chosen_)
        np.save(filename + 'mi' + t + str(i) + "_chosen.npy", chosen)
        np.save(filename + 'mi' + t + str(i) + "_lines.npy", lines)
        np.save(filename + 'mi' + t + str(i) + "_newlines.npy", newlines)

        # Fit GP
        if i % 2 == 0:
            compute_error(lines, functions, K, i, t, filename, S, Xi, Xj)




def mi2(t, S_points, E_points,  functions, S, Xi, Xj, filename, maxsensors):
    K = Numeric(lengthscale=15., variance=0.5, noise=0.001, kerneltype='Matern32', period=period, samples=10)
    criterion = MI_ultrasound(t, K, second=True)
    chosen_, E_points, lines_ = criterion.initialize(S_points, E_points)
    errors = []
    chosen = []
    lines = []
    scores = np.ones(len(E_points)) * 100000
    scores_updated = np.zeros(len(E_points))
    chosen.append(chosen_)
    lines.extend(lines_)

    for i in range(maxsensors - 1):
        print(i)
        sys.stdout.flush()
        chosen_, newlines, E_points, scores_updated, scores = criterion.get_nextlines(S_points, E_points, lines, chosen,
                                                                                      scores_updated, scores)
        lines.extend(newlines)
        chosen.append(chosen_)
        np.save(filename + 'mi' + t + str(i) + "_chosen.npy", chosen)
        np.save(filename + 'mi' + t + str(i) + "_lines.npy", lines)
        np.save(filename + 'mi' + t + str(i) + "_newlines.npy", newlines)

        # Fit GP
        if i % 2 == 0:
            compute_error(lines, functions, K, i, t, filename, S, Xi, Xj)


if __name__ == "__main__":

    period = 2 * np.pi * 35

    f1 = lambda xi, xj: 1 if 120 <= xj <= 170 and -55 <= xi <= -25 else 0
    f2 = lambda xi, xj: 1 if 150 <= xj <= 200 and 60 <= xi <= 110 else 0
    f3 = lambda xi, xj: 1 if 150 <= xj <= 230 and -100 <= xi <= -70 else 0
    f4 = lambda xi, xj: 1 if 200 <= xj <= 240 and -40 <= xi <= 40 else 0
    f5 = lambda xi, xj: 1 if 60 <= xj <= 240 and -10 <= xi <= 20 else 0
    f6 = lambda xi, xj: 1 if 130 <= xj <= 150 and -100 <= xi <= 100 else 0
    f7 = lambda xi, xj: 1 if 50 <= xj <= 100 and -110 <= xi <= -60 else 0
    f8 = lambda xi, xj: 1 if 150 <= xj <= 230 and 70 <= xi <= 110 else 0
    f9 = lambda xi, xj: 1 if 100 <= xj <= 140 and -40 <= xi <= 40 else 0
    f10 = lambda xi, xj: 1 if 60 <= xj <= 240 and -50 <= xi <= 50 else 0

    functions = [f1, f2, f3, f4, f5, f6, f7, f8, f9, f10]


    def value(xi, xj, func):
        if xi < -period / 2:
            xi += period
        if xi > period / 2:
            xi -= period
        if xi < -period / 2:
            xi += period
        if xi > period / 2:
            xi -= period
        return func(xi, xj)


    n_x = 100
    n_y = 130
    gridstart_x = -period / 2
    gridend_x = period / 2
    gridstart_y = 0
    gridend_y = 300
    second = True
    third = False

    n_start_x = 5
    n_start_y = 3
    # V_start = np.array([[0,0]])
    V_start = np.array(np.meshgrid(np.linspace(gridstart_x + 10, gridend_x - 10, n_start_x),
                                   np.linspace(gridstart_y, gridstart_y + 100, n_start_y)))
    V_start = V_start.reshape(2, n_start_x * n_start_y).T

    n_end_x = FLAGS.xgrid
    n_end_y = FLAGS.ygrid
    maxsensors = FLAGS.maxsensors
    V_end = np.array(np.meshgrid(np.linspace(gridstart_x + 10, gridend_x - 10, n_end_x),
                                 np.linspace(gridstart_y + 150, gridend_y, n_end_y)))
    V_end = V_end.reshape(2, n_end_x * n_end_y).T


    all_lines = []
    for point1 in V_start:
        for point2 in V_end:
            if not np.array_equal(point1, point2):
                all_lines.append([point1, point2])
                if second == True:
                    all_lines.append([point1, point2 + [period, 0]])
                    all_lines.append([point1, point2 + [-period, 0]])
                if third == True:
                    all_lines.append([point1, point2 + [2 * period, 0]])
                    all_lines.append([point1, point2 + [-2 * period, 0]])
    all_lines = np.array(all_lines)

    n = 50
    [Xi, Xj] = np.meshgrid(np.linspace(-110, 110, n), np.linspace(0, 300, n))
    S = np.vstack((Xi.ravel(), Xj.ravel())).T



    criter = FLAGS.criter
    t = FLAGS.summarytype

    if criter == 'entropy':
        print("entropy")
        filename = "data/ultrasound/toy_bothgrid_entropy_" + str(n_end_x)  + str(n_end_y)
        entropy(t, V_start, V_end, functions, S, Xi, Xj, filename, maxsensors)

    if criter == 'mi':
        print("mi")
        print(t)
        filename = "data/ultrasound/toy_both_mi_" + str(n_end_x)  + str(n_end_y)
        mi(t, V_start, V_end, functions, S, Xi, Xj, filename, maxsensors)


